==> multiphasepy-1.3.0.tar.gz: multiphasepy-1.3.0/multiphasepy/foam.py <==
# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for OpenFOAM functionality.

This module contains functions to operate on a OpenFOAM case, for example
reading of different OpenFOAM output file types or checking for a valid
OpenFOAM case.

Attributes:
    csvdelimiter (str): Delimiter for OpenFOAM comma separated files, which is
        supposed to be a comma, but in OpenFOAM it is a whitespace
    comments (str): Character indicating a comment line in OpenFOAM files
    filesuffix (str): General suffix in OpenFOAM for dictionary files
    indention(int): Number of whitespace used of indention
"""

from sys import exit
import csv
import linecache
import logging
import yaml
from os import path, makedirs
from typing import Iterable, Tuple, Sequence, Any, NoReturn
from pygments.lexers import CppLexer
from pygments.token import Name, Text, Comment, Punctuation, Keyword
from pygments.filter import Filter
from datetime import date
from ._version import __version__
import numpy as np
import re

# Enable logging for this module
log = logging.getLogger(__name__)

csvdelimiter = " "
comments = "#"
filesuffix = ".orig"
indention = 4


class filterdictionary(Filter):
    """A custom filter for configuration files."""

    def __init__(self, **options):
        """Initialize filter class."""
        Filter.__init__(self, **options)

    def filter(self, lexer, stream):
        """Filter to remove or convert tokens.

        The configuration files are only similar to C++. This filter function
        removes the unnecessary tokens and converts C++ specific tokens into
        more simpler tokens.

        Args:
            lexer: lexer used to tokenize stream.
            stream: stream to be tokenized.

        Returns:
            A filtered tokenized stream.
        """
        for t, v in stream:
            if t is Text.Whitespace:
                continue
            elif t is Comment.Single:
                continue
            elif t is Comment.Multiline:
                continue
            elif t is Keyword:
                t = Name
            yield t, v


def read_data(fname: str) -> Tuple[np.ndarray, Sequence[str]]:
    """Read simulation data.

    This function reads data from simulations. Depending on the file ending
    a different reader is invoked.

    Args:
        fname: File name of the data file to be read.

    Returns:
        A tuple of the column names and a numpy data array.
    """
    if path.isfile(fname):
        _, ftype = path.splitext(fname)

        if ftype == ".csv":
            log.debug("Loading file with csv.DictReader.")
            with open(fname, "r", newline="", encoding="utf-8") as f:
                d_reader = csv.DictReader(f, delimiter=csvdelimiter)
                header = d_reader.fieldnames
                data = np.asarray(list(d_reader.reader), dtype=np.float64)
        elif ftype == ".dat":
            log.debug("Loading text file with numpy.genfromtxt.")
            hlines = 0
            with open(fname, encoding="utf-8") as f:
                while True:
                    line = f.readline()
                    if line.startswith("#"):
                        hlines += 1
                    else:
                        break
            # Assume that the last header lines contains column headers
            header = linecache.getline(fname, hlines).strip(comments).split()
            data = np.genfromtxt(
                fname, comments=comments, encoding="utf-8", dtype=np.float64
            )

            if len(header) != data.shape[1]:
                log.error("Column header do not match the data array.")
                log.debug(f"Header has {len(header)} entries: {header}")
                log.debug(f"Data array has {data.shape[1]} columns")
                exit(1)
        elif is_probes_file(fname):
            log.debug("Loading probes file from OpenFOAM Foundation software.")
            data, header = get_probes_data(fname)
        else:
            log.error(f"Unknown file format {ftype}")
            exit(1)

        return data, header
    else:
        log.error(f"File {fname} does not exist.")
        exit(1)


def write_data(
    fname: str,
    data: Iterable[Iterable[Any]],
    header: Iterable[Any] = "",
    fmode: str = "w",
) -> NoReturn:
    """Write data.

    This function write data. Depending on the file ending a different writer is
    invoked.

    Args:
        data: the data to be written to file
        fname: the file name of the output data file
        fmode: the mode of the output file('w' for new file and 'x' for append)
        header: field names to be written as header

    Returns:
        Nothing is returned.
    """
    if path.dirname(fname) != "":
        makedirs(path.dirname(fname), exist_ok=True)
    _, ftype = path.splitext(fname)

    if ftype == ".csv":
        log.debug("Writing a csv file with csv.writer.")
        with open(fname, fmode, newline="", encoding="utf-8") as f:
            writer = csv.writer(f, delimiter=" ")
            writer.writerow(header)
            writer.writerows(data)
    elif ftype == ".dat":
        log.debug("Writing a text file with numpy.savetxt.")
        np.savetxt(
            fname,
            data,
            fmt="%.6e",
            delimiter=csvdelimiter,
            newline="\n",
            header=csvdelimiter.join(header),
            comments=comments,
            encoding="utf-8",
        )
    else:
        log.error(f"Unknown file format {ftype}")
        exit(1)


def foamtodict(stream: Iterable, depth: int = 0) -> dict:
    """Convert a tokenized stream into a dictionary.

    This function converts a stream tokenized with cpplexer from a configuration
    file, e.g., controlDict into a python dictionary. The function calls itself
    if a sub-dictionary found.

    Args:
        stream: tokenized input stream
        depth: variable to check depth of dictionaries

    Returns:
        Dictionary with settings specified in configuration file.
    """
    d = {}
    depth = depth + 1
    for t, v in stream:
        # First token has to be a named key
        if t is Name:
            (tt, vv) = next(stream)
            # Token indicates a new dictionary
            if tt is Punctuation and vv == "{":
                d[v] = foamtodict(stream, depth)
            else:
                # Token indicates a new list
                if tt is Punctuation and vv == "(":
                    d[v] = foamtolist(stream)
                # Token indicates a value
                else:
                    d[v] = vv
                # Check for proper punctuation token after list and values
                (_, vvv) = next(stream, (Punctuation, "EOF"))
                if vvv != ";":
                    log.error(f"Missing ; or ) in line: {v} {vv}")
                    exit(1)
        # Token indicates end of dictionary
        elif t is Punctuation and v == "}":
            depth = depth - 1
            break
        # Found some not defined token
        else:
            log.error(f"Something went wrong for type {t} with value {v}.")
            exit(1)
    if depth != 1:
        log.error("Missing } to close dictionary.")
        exit(1)
    return d


def foamtolist(stream: Iterable) -> list:
    """Convert a tokenized stream into a list.

    This function converts a stream tokenized with cpplexer from a configuration
    file, e.g., controlDict into a python list.

    Args:
        stream: tokenized input stream

    Returns:
        List with values specified in configuration file.
    """
    li = []
    while True:
        (t, v) = next(stream)
        if t == Punctuation:
            break
        li.append(v)
    return li


def dicttofoam(dictionary: dict, depth: int = 0) -> list:
    """Convert a dictionary to configuration file.

    This function converts a dictionary to a configuration file for OpenFOAM
    Foundation software. This is a recursive function, which calls it self
    for any dictionary found in the given dictionary.

    Args:
        dictionary: dictionary with configuration data
        depth: dictionary depth, mainly used to indent entries

    Returns:
        List with formated lines for the configuration file.
    """
    lines = []
    for key, value in dictionary.items():
        if type(value) is dict:
            lines.append(f"{' ' * depth * indention}{key}")
            lines.append("{")
            result = dicttofoam(value, depth=depth + 1)
            lines += result
            lines.append(f"{' ' * (depth - 1) * indention}}}")
        elif type(value) is list:
            lines.append(f"{' ' * depth * indention}{key} {listtofoam(value)}")
        else:
            lines.append(f"{' ' * depth * indention}{key} {value};")
    return lines


def listtofoam(pylist: list) -> str:
    """Convert a list to configuration file.

    This function converts a list to a entry for a configuration file for
    OpenFOAM Foundation software.

    Args:
        pylist: list with configuration data

    Returns:
        Formated line for the configuration file.
    """
    return f"({' '.join(pylist)});"


def read_dict(fname: str) -> dict:
    """Read configuration file into dictionary.

    This function opens a configuration file and converts it to a dictionary.
    For dictionaries read from OpenFOAM Foundation software the input is
    tokenized with the CppLexer and converted to a dictionary. The input is also
    filtered to remove unnecessary content.

    Args:
        fname: configuration file

    Returns:
        Dictionary with settings specified in configuration file.
    """
    if path.isfile(fname):
        _, ftype = path.splitext(fname)

    if ftype == ".orig" or path.basename(fname).endswith("Dict"):
        log.debug("Loading dictionary from OpenFOAM Foundation software.")
        with open(fname, "r", encoding="utf-8") as f:
            lexer = CppLexer(stripall=True)
            lexer.add_filter(filterdictionary())
            return foamtodict(lexer.get_tokens(f.read()))
    elif ftype == ".yml" or ftype == ".yaml":
        log.debug("Loading dictionary with yaml.safe_load.")
        with open(fname, "r", encoding="utf-8") as f:
            return yaml.safe_load(f)

    else:
        log.error(f"Unknown file format {ftype}")
        exit(1)

    log.error(f"File {fname} does not exist")
    exit(1)


def write_dict(
    dictionary: dict,
    fname: str,
    fmode: str = "w",
) -> NoReturn:
    """Write dictionaries to configuration files.

    This function writes dictionaries to configuration files. Depending on the
    file ending a different writer is invoked.

    Args:
        dictionary: the dictionary to be written to file
        fname: the file name of the output configuration file
        fmode: the mode of the output file('w' for new file and 'x' for append)

    Returns:
        Nothing is returned.
    """
    makedirs(path.dirname(fname), exist_ok=True)
    _, ftype = path.splitext(fname)

    if path.basename(fname).endswith("Dict"):
        lines = [
            f"/* Generated automatically by multiphasepy version: \
{__version__} at: {date.today():%Y-%m-%d} */"
        ]
        lines += dicttofoam(dictionary)
        with open(fname, "w") as f:
            f.write("\n".join(lines))
    elif ftype == ".yml" or ftype == ".yaml":
        with open(fname, fmode, encoding="utf-8") as f:
            yaml.dump(dictionary, f)
    else:
        log.error(f"Unknown file format {ftype}")
        exit(1)


def getfilename(fname: str) -> str:
    """Check for orig files.

    The simulation software uses the .orig suffix to identify original files
    that get processed into a new file without the .orig file extension.
    This function checks if for a given filename the file exists, in case not
    it checks for the same filename but with .orig suffix.

    Args:
        fname: path to simulation configuration file

    Returns:
        Returns the file name with or without suffix or None if both files are
        not found.
    """
    if path.isfile(fname):
        return fname
    elif path.isfile(fname + filesuffix):
        return fname + filesuffix
    else:
        log.error(f"Found neither {fname} nor {fname + filesuffix}.")
        exit(1)


def convert(infile: str, outfile: str) -> NoReturn:
    """Converts a given data file.

    This function converts a given datafile typically written by the simulation
    software to an output file.

    Args:
        infile: input file
        outfile: output file

    Returns:
        Nothing is returned.
    """
    write_data(path.abspath(outfile), *(read_data(path.abspath(infile))))


def is_case(dname: str) -> bool:
    """Check for simulation software case.

    This function checks if the directory contains a regular simulation software
    case, by testing for system/controlDict or system/controlDict.orig.

    Args:
        dname: directory name of the OpenFOAM case

    Returns:
        True if it is a regular OpenFOAM case, false if not.
    """
    return path.isdir(dname) and (
        path.isfile(path.join(dname, "system", "controlDict"))
        or path.isfile(path.join(dname, "system", "controlDict" + filesuffix))
    )


def is_probes_file(fname: str) -> bool:
    """Test for a probe data file.

    The test is based on finding the word 'Probe' somewhere in the header.

    Args:
        fname: file name containing probe data

    Returns:
        True if it is a probe file.
    """
    with open(fname, encoding="utf-8") as f:
        while True:
            line = f.readline()
            if line.startswith("#"):
                if line.find("Probe"):
                    return True
            else:
                break

    return False


def get_probes_data(fname: str) -> Tuple[np.ndarray, Sequence[str]]:
    """Read and convert a probe data file.

    Args:
        fname: file name containing probe data

    Returns:
        A tuple of column names and a numpy data array with validation
        data.
    """
    hlines = 0
    probes = []
    proberegex = re.compile(
        "^#\\sProbe\\s\\d\\s(\\((\\d*\\.?\\d*\\s){2}\\d*\\.?\\d*\\))$"
    )

    with open(fname, encoding="utf-8") as f:
        while True:
            line = f.readline()
            if line.startswith("#"):
                hlines += 1
                p = proberegex.match(line)
                if p is not None:
                    probes.append(p.group(1))
            else:
                break

    firstdataline = linecache.getline(fname, hlines + 1).split()
    probeheader = ["Time"]
    i = 1
    while i < len(firstdataline):
        if firstdataline[i].startswith("("):
            idx = int((i - 1) / 3)
            probeheader.append(f"Probe {idx} {probes[idx]} X")
            probeheader.append(f"Probe {idx} {probes[idx]} Y")
            probeheader.append(f"Probe {idx} {probes[idx]} Z")
            i += 3
        else:
            probeheader.append(f"Probe {i - 1} {probes[i - 1]}")
            i += 1

    # Remove parentheses, apply to all available columns
    convs = {
        k: lambda s: np.float64(s.strip("()"))
        for k in range(0, len(firstdataline))
    }

    # Save data to numpy array
    probedata = np.genfromtxt(
        fname,
        comments=comments,
        encoding="utf-8",
        dtype=np.float64,
        converters=convs,
    )

    return probedata, probeheader

==> multiphasepy-1.4.5.tar.gz: multiphasepy-1.4.5/multiphasepy/foam.py <==
# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for OpenFOAM functionality.

This module contains functions to operate on a OpenFOAM case, for example
reading of different OpenFOAM output file types or checking for a valid
OpenFOAM case.

Attributes:
    csvdelimiter (str): Delimiter for OpenFOAM comma separated files, which is
        supposed to be a comma, but in OpenFOAM it is a whitespace
    comments (str): Character indicating a comment line in OpenFOAM files
    filesuffix (str): General suffix in OpenFOAM for dictionary files
    indentation(int): Number of whitespace used of indentation
"""

import logging
from os import path
from typing import Iterable

from pygments.filter import Filter
from pygments.lexers import CppLexer
from pygments.token import Comment, Keyword, Name, Punctuation, Text

# Enable logging for this module
log = logging.getLogger(__name__)

csvdelimiter = " "
comments = "#"
filesuffix = ".orig"
indentation = 4


class filterdictionary(Filter):
    """A custom filter for configuration files."""

    def __init__(self, **options):
        """Initialize filter class."""
        Filter.__init__(self, **options)

    def filter(self, lexer, stream):
        """Filter to remove or convert tokens.

        The configuration files are only similar to C++. This filter function
        removes the unnecessary tokens and converts C++ specific tokens into
        more simpler tokens.

        Args:
            lexer: lexer used to tokenize stream.
            stream: stream to be tokenized.

        Returns:
            A filtered tokenized stream.
        """
        for t, v in stream:
            if t is Text.Whitespace:
                continue
            elif t is Comment.Single:
                continue
            elif t is Comment.Multiline:
                continue
            elif t is Keyword:
                t = Name
            elif t is Name.Constant:
                t = Name
            yield t, v


def getdict(fname: str) -> dict:
    """Read a dictionary into a python dictionary.

    This function reads a dictionary from the OpenFOAM Foundation software
    into a python dictionary to allow manipulation with python. The stream
    is tokenized with cpplexer, as the syntax in the OpenFOAM Foundation
    software is very similar to C++ syntax.

    Args:
        fname: file name containing the dictionary.

    Returns:
        A python dictionary.
    """
    with open(fname, "r", encoding="utf-8") as f:
        lexer = CppLexer(stripall=True)
        lexer.add_filter(filterdictionary())
        return foamtodict(lexer.get_tokens(f.read()))


def foamtodict(stream: Iterable, depth: int = 0) -> dict:
    """Convert a tokenized stream into a dictionary.

    This function converts a stream tokenized with cpplexer from a configuration
    file, e.g., controlDict into a python dictionary. The function calls itself
    if a sub-dictionary found.

    Args:
        stream: tokenized input stream
        depth: variable to check depth of dictionaries

    Returns:
        Dictionary with settings specified in configuration file.
    """
    d = {}
    depth = depth + 1
    for t, v in stream:
        # First token has to be a named key
        if t is Name:
            (tt, vv) = next(stream)
            # Token indicates a new dictionary
            if tt is Punctuation and vv == "{":
                d[v] = foamtodict(stream, depth)
            else:
                # Token indicates a new list
                if tt is Punctuation and vv == "(":
                    d[v] = foamtolist(stream)
                # Token indicates a value
                else:
                    d[v] = vv
                # Check for proper punctuation token after list and values
                (_, vvv) = next(stream, (Punctuation, "EOF"))
                if vvv != ";":
                    raise ValueError(f"Missing ; or ) in line: {v} {vv}")
        # Token indicates end of dictionary
        elif t is Punctuation and v == "}":
            depth = depth - 1
            break
        # Found some not defined token
        else:
            raise ValueError(
                f"Something went wrong for type {t} with value {v}."
            )
    if depth != 1:
        raise ValueError("Missing } to close dictionary.")
    return d


def foamtolist(stream: Iterable) -> list:
    """Convert a tokenized stream into a list.

    This function converts a stream tokenized with cpplexer from a configuration
    file, e.g., controlDict into a python list.

    Args:
        stream: tokenized input stream

    Returns:
        List with values specified in configuration file.
    """
    li = []
    while True:
        (t, v) = next(stream)
        if t == Punctuation:
            break
        li.append(v)
    return li


def dicttofoam(dictionary: dict, depth: int = 0) -> list:
    """Convert a dictionary to configuration file.

    This function converts a dictionary to a configuration file for OpenFOAM
    Foundation software. This is a recursive function, which calls it self
    for any dictionary found in the given dictionary.

    Args:
        dictionary: dictionary with configuration data
        depth: dictionary depth, mainly used to indent entries

    Returns:
        List with formatted lines for the configuration file.
    """
    lines = []
    for key, value in dictionary.items():
        if type(value) is dict:
            lines.append(f"{' ' * depth * indentation}{key}")
            lines.append("{")
            result = dicttofoam(value, depth=depth + 1)
            lines += result
            lines.append(f"{' ' * (depth - 1) * indentation}}}")
        elif type(value) is list:
            lines.append(
                f"{' ' * depth * indentation}{key} {listtofoam(value)}"
            )
        else:
            lines.append(f"{' ' * depth * indentation}{key} {value};")
    return lines


def listtofoam(pylist: list) -> str:
    """Convert a list to configuration file.

    This function converts a list to a entry for a configuration file for
    OpenFOAM Foundation software.

    Args:
        pylist: list with configuration data

    Returns:
        formatted line for the configuration file.
    """
    return f"({' '.join(pylist)});"


def getfilename(fname: str) -> str:
    """Check for orig files.

    The simulation software uses the .orig suffix to identify original files
    that get processed into a new file without the .orig file extension.
    This function checks if for a given filename the file exists, in case not
    it checks for the same filename but with .orig suffix.

    Args:
        fname: path to simulation configuration file

    Returns:
        Returns the file name with or without suffix or None if both files are
        not found.
    """
    if path.isfile(fname):
        return fname
    elif path.isfile(fname + filesuffix):
        return fname + filesuffix
    else:
        raise FileNotFoundError(
            f"Found neither {fname} nor {fname + filesuffix}."
        )


def is_case(dname: str) -> bool:
    """Check for simulation software case.

    This function checks if the directory contains a regular simulation software
    case, by testing for system/controlDict or system/controlDict.orig.

    Args:
        dname: directory name of the OpenFOAM case

    Returns:
        True if it is a regular OpenFOAM case, false if not.
    """
    return path.isdir(dname) and (
        path.isfile(path.join(dname, "system", "controlDict"))
        or path.isfile(path.join(dname, "system", "controlDict" + filesuffix))
    )


def is_probes_file(fname: str) -> bool:
    """Test for a probe data file.

    The test is based on finding the word 'Probe' somewhere in the header.

    Args:
        fname: file name containing probe data

    Returns:
        True if it is a probe file.
    """
    with open(fname, encoding="utf-8") as f:
        while True:
            line = f.readline()
            if line.startswith("#"):
                if line.find("Probe"):
                    return True
            else:
                break

    return False
